#pragma module BCDRIVER "LMS version 2.00"
 
/*
 * CONFIGURATION INFORMATION
 * 
 *   System type - AlphaServer 800 5/500
 *   Interface   - bc635PCI Datum Time Frequency Processor
 *   Bus         - PCI
 *   TR number   - 2
 *   Slot number - 12
 *   NODE        - 96 (0x60)
 *   Class       - DC$_REALTIME
 *   Device type - DT$_USER9
 *   buffer size - 128
 *   interrupt   - PCI interrupt pin #INTA
 *
 */

/* Define system data structure types and constants */

#include <bufiodef.h>		/* Define the packet header for a system */
				/*   buffer for buffered I/O data */
#include <adpdef.h>             /* Adapter control block */
#include <ccbdef.h>             /* Channel control block */
#include <crbdef.h>             /* Controller request block */
#include <cramdef.h>            /* Controller register access method */
#include <dcdef.h>              /* Device codes */
#include <ddbdef.h>             /* Device data block */
#include <ddtdef.h>             /* Driver dispatch table */
#include <devdef.h>             /* Device characteristics */
#include <dptdef.h>             /* Driver prologue table */
#include <fdtdef.h>             /* Function decision table */
#include <fkbdef.h>             /* Fork block */
#include <hwrpbdef.h>           /* Hardware restart parameter block */
#include <idbdef.h>             /* Interrupt data block */
#include <ints.h>               /* Integer typedefs     */
#include <iocdef.h>             /* IOC constants        */
#include <iodef.h>              /* I/O function codes */
#include <irpdef.h>             /* I/O request packet */
#include <ka0602def.h>          /* DEC 2000 Model 300 AXP specific defs  */
#include <lpdef.h>              /* Line printer definitions */
#include <orbdef.h>             /* Object rights block */
#include <pcidef.h>             /* PCI bus configuration definitions */
#include <pcbdef.h>             /* Process control block             */
#include <opcdef.h>             /* OPCOM message definitions         */
#include <msgdef.h>             /* System-wide mailbox message codes */
#include <ssdef.h>              /* System service status codes       */
#include <stsdef.h>             /* Status value fields */
#include <ucbdef.h>             /* Unit control block */
#include <vecdef.h>             /* IDB interrupt transfer vector */
#include <mcjdef.h>             /* The Magic Cookie Jar structure */

/* Define function prototypes for system routines */

#include <exe_routines.h>       /* Prototypes for exe$ and exe_std$ routines */
#include <ioc_routines.h>       /* Prototypes for ioc$ and ioc_std$ routines */
#include <sch_routines.h>       /* Prototypes for sch$ and sch_std$ routines */

/* Define various device driver macros */

#include <vms_drivers.h>        /* Device driver support macros, including */
                                /* table initialization macros and prototypes*/

/* Define the DEC C functions used by this driver */

#include <builtins.h>           /* OpenVMS AXP specific C builtin functions */
#include <string.h>             /* String routines provided by "kernel CRTL" */


#include "src$:bcdriver.h"      /* Definitions shared with applications */


/*****************************************************************************/
/*                                                                           */
/* Define constants specific to this driver                                  */
/*                                                                           */
/*****************************************************************************/


#define USER_READ_BUFFER 128 /* The user must specify a 128 byte read buffer */
enum {                          /* Miscellaneous constants */
    FALSE              =   0,   /* True and False Flags */
    TRUE               =   1,
    DEVICE_IPL         =  21,   /*   Interrupt priority level of device */
    NUMBER_CRAMS       =   0,   /*   Number of CRAMs needed */
    DATA_EXPND_CUSHION =  32    /*   Extra room in system buffer for expansion */
};

enum {                          /* Define various timeout constants */
    BC_WFI_TMO     =       2    /*   Interrupt timeout value in seconds */
};


/*
 * Define the bc635 device register offsets.
 *
 *    ACCESS TYPE   A   -- Just access to trigger function (should use write)
 *                  R   -- Read to access information
 *                  W   -- Write to set information 
 *                  R/W -- Read and write to access/set information
 *
 *    RESET STATE   U   -- State after reset is undefined
 *                  0   -- State after reset is all zeros
 */
                             /* TYPE RESET DESCRIPTION             */
#define BC_DR_TIMEREQ  0x0000  /*   A    U   Time request            */
#define BC_DR_EVENTRE  0x0004  /*   A    U   Event request           */
#define BC_DR_UNLOCK   0x0008  /*   A    U   Release capture lockout */
/*                     0x000C  NOT USED */
#define BC_DR_CONTROL  0x0010  /*  R/W   0   Control register        */
#define BC_DR_ACK      0x0014  /*  R/W   U   Acknowledge register    */
#define BC_DR_MASK     0x0018  /*  R/W   0   Interrupt mask          */
#define BC_DR_INTSTAT  0x001C  /*  R/W   0   Interrupt status        */
#define BC_DR_MINSTRB  0x0020  /*   W    U   Minor strobe time       */
#define BC_DR_MAJSTRB  0x0024  /*   W    U   Major strobe time       */
/*                     0x0028  NOT USED */
/*                     0x002C  NOT USED */
#define BC_DR_TIME0    0x0030  /*   R    U   Minor time holding reg. */
#define BC_DR_TIME1    0x0034  /*   R    U   Major time holding reg. */
#define BC_DR_EVENT0   0x0038  /*   R    U   Minor event holding reg.*/
#define BC_DR_EVENT1   0x003C  /*   R    U   Major event holding reg.*/

/*
 * This is the offset to base of the dual ported RAM area from the value in
 * PCI configuration base address 1. This skips over the space that contains
 * the bc635PCI ROM.
 */
#define BC_ROM_SIZE    0x0800

/*
 * These are the offset to dual ported RAM areas that contain the offsets
 * to the put (write, input areas from the bc635PCI view) and the get areas.
 */
#define BC_TOP         0x0FFF
#define BC_TOP_1       0x0FFE
#define BC_TOP_2       0x0FFD
#define BC_TOP_3       0X0FFC
#define BC_TOP_4       0X0FFB
#define BC_TOP_5       0X0FFA
#define BC_TOP_6       0X0FF9
#define BC_TOP_7       0X0FF8

/* Time Frequency Processor (TFP) Control Register
 * Mask values are defined for each of the control bits in the TFP.  This
 * driver always writes a new value to the TFP when a bit needs to be set.
 * A convenient way of doing this is to logically or together a subset of the
 * following masks to form the new LPC value.
 */
enum tfpc_masks {
    TFP_C_LOCKEN   = 0x01,  /* Event capture lockout enable */
    TFP_C_EVSOURCE = 0x02,  /* Event time capture source select */
    TFP_C_EVSENSE  = 0x04,  /* Event input edge select */
    TFP_C_EVENTEN  = 0x08,  /* Event capture enable */
    TFP_C_STREN    = 0x10,  /* Time coincidence strobe output enable */
    TFP_C_STRMODE  = 0x20,  /* Time coincidence strobe mode */
    TFP_C_FREQSEL0 = 0x40,  /* Output frequency select */
    TFP_C_FREQSEL1 = 0x80   /* Output frequency select */
                };
/*
 * TFP ACK register bits. This register controls access to TFP memory shared
 * by the TFP and the computer.
 */
enum tfpa_bits {
    TFP_A_CMDIN  = 0x01,  /* Set when command has been read by the TFP */
    TFP_A_GPSIN  = 0x02,  /* Set when the TFP has a new GPS packet */
    TFP_A_CMDRDY = 0x80   /* Set when CPU has written command for the TFP */
                };
/*
 * TFP interrupt bits. These bits are used to enable interrupts in the MASK
 * register and to check which interrupt(s) occured in the INTSTAT register.
 */
enum tfpi_bits {
    TFP_I_EVENT    = 0x01,  /* Event Input event occurance */
    TFP_I_PERIODIC = 0x02,  /* Periodic output occurance */
    TFP_I_STROBE   = 0x04,  /* Time coincidence strobe occurance */
    TFP_I_SECOND   = 0x08,  /* One second epoch (1 PPS) occurance */
    TFP_I_GPS      = 0x10   /* GPS DATA PACKED AVAILABLE (bc637PCI only) */
                };
#define TFP_I_CLEAR 0x1F    /* Used to clear all pending interrupt bits */

/*
 * TFP status update/read indicator values
 */
#define TFP_NO_ACTION 0x00
#define TFP_LOAD      0x01
#define TFP_READ      0x02

/*
 * TFP DPRAM command defintions
 */
#define TFP_CMD_SELECT_MODE 0x10 /* Select TFP timing mode (reset MODE 0) */
#define TFP_CMD_SET_TIME    0x12 /* Set Major time (reset N/A) */
#define TFP_CMD_SET_YEAR    0x13 /* Set year (reset 0) */
#define TFP_CMD_SET_OUTPUT  0x14 /* Set periodic output (reset N/A) */
#define TFP_CMD_SET_FORMAT  0x15 /* Select time code format (reset 'B') */
#define TFP_CMD_SET_TYPE    0x16 /* Select time code modulation type (reset M)*/
#define TFP_CMD_SET_OFFSET  0x17 /* Set timing offset (reset 0) */
#define TFP_CMD_REQUEST     0x19 /* Request data from TFP (reset N/A) */
#define TFP_CMD_RESET       0x1A /* Software reset (reset N/A) */
#define TFP_CMD_SET_SOURCE  0x20 /* Select clock source (reset 1) */
#define TFP_CMD_CONTROL_JAM 0x21 /* Control JAM-SYNC (reset 1) */
#define TFP_CMD_FORCE_JAM   0x22 /* Force JAM-SYNC (reset N/A) */
#define TFP_CMD_CONTROL_CLK 0x23 /* Control clock disciplining (reset 1) */
#define TFP_CMD_LOAD        0x24 /* Load D/A converter (reset N/A) */
#define TFP_CMD_SET_GAIN    0x25 /* Set discipline gain (reset N/A) */
#define TFP_CMD_SEND        0x30 /* Send packet to GPS receiver (reset N/A) */
#define TFP_CMD_REQUEST_GPS 0x31 /* Request packet from GPS receiver (reset N/A) */
#define TFP_CMD_MANUAL      0x32 /* Manual request packet from GPS receiver (reset N/A) */
#define TFP_CMD_SET_GPS     0x33 /* Select GPS time format (reset 0) */

/*                                                                            */
/* Define Device-Dependent Unit Control Block with extensions for BC device   */
/* The extensions maintain the current values read from or set to the device  */
/*                                                                            */
/* NOTE: Currently there is no need for most of the information from the IRIG */
/*       board and is is not read. See interrupt service routine for the data */
/*       that is actually being read back. Room has been left in the data     */
/*       structure to handle a read back of everything, but this would take   */
/*       additional time at the device driver IPL level.                      */
/*                                                                            */

typedef struct {
    UCB    ucb$r_ucb;                   /* Generic UCB */
    uint64 ucb$l_bc_iohandle;           /* DPRAM IO handle for bc635PCI */
    uint64 ucb$l_bc_reghandle;          /* Register IO handle for bc635PCI */
    uint64 ucb$l_bc_counter;     /* (Live) Count of expected interrupts */
    uint64 ucb$l_bc_extra;       /* (Live) Count of unexpected interrupts */
    int    ucb$l_bc_status;             /* Request/set indicator for TFP */
    int    ucb$l_bc_time1;       /* (Live) TIME 1, TFP Binary time */
    int    ucb$l_bc_time0;       /* (Live) TIME 0, TFP Binary time */
    int    ucb$l_bc_mode;               /* Timing mode */
    int    ucb$l_bc_format;             /* Timing format */
    int    ucb$l_bc_year;        /* (Live) Year 1970 -- 2050 */
    int    ucb$l_bc_sync;               /* Sync to 1 pps flag */
    int    ucb$l_bc_n1;                 /* Sync divider 1 */
    int    ucb$l_bc_n2;                 /* Sync divider 2 */
    int    ucb$l_bc_code;               /* Time code format */
    int    ucb$l_bc_modulation;         /* Time code modulation */
    int    ucb$l_bc_offset;      /* (Live) timing offset */
    int    ucb$l_bc_utc;                /* UTC support flag */
    int    ucb$l_bc_leap_second;        /* Leap second count */
    int    ucb$l_bc_leap_event;         /* Leap event direction */
    int    ucb$l_bc_leap_event_time;    /* Leap event time */
    int    ucb$l_bc_fw_major;           /* Firmware major version number */
    int    ucb$l_bc_fw_minor;           /* Firmware minor version number */
    int    ucb$l_bc_fw_month;           /* Firmware month */
    int    ucb$l_bc_fw_day;             /* Firmware day */
    int    ucb$l_bc_fw_year;            /* Firmware year */
    int    ucb$l_bc_source;             /* Clock source */
    int    ucb$l_bc_jam;                /* JAM synch control flag */
    int    ucb$l_bc_disciplining;       /* Disciplining control */
    int    ucb$l_bc_phase;              /* Phase control flag */
    int    ucb$l_bc_step;               /* Phase step */
    int    ucb$l_bc_gain;               /* Disciplining gain */
    int    ucb$l_bc_filter;             /* Filter constant */
    int    ucb$l_bc_value;              /* D/A value */
    int    ucb$l_bc_register;           /* Data to move to/from register */
    int    ucb$l_bc_year_offset; /* (Live) Offset to year area of RAM */
    int    ucb$l_bc_gps_offset;  /* (Live) Offset to GPS MSG area of RAM */
    int    ucb$l_bc_get_offset;  /* (Live) Offset to output area of RAM */
    int    ucb$l_bc_put_offset;  /* (Live) Offset to input area of RAM */
    uint8  ucb$l_bc_response[24];/* (Live) Data read from dual port RAM */
    } BC_UCB;

/* Prototypes for driver routines defined in this module */

/* Driver table initialization routine */

    int  driver$init_tables ();

/* Device I/O database structure initialization routine */

    void bc$struc_init (CRB *crb, DDB *ddb, IDB *idb, ORB *orb, BC_UCB *ucb);

/* Device I/O database structure re-initialization routine */

    void bc$struc_reinit (CRB *crb, DDB *ddb, IDB *idb, ORB *orb, BC_UCB *ucb);

/* Controller initialization routine */

    int  bc$ctrl_init (IDB *idb, DDB *ddb, CRB *crb);

/* Unit initialization routine */

    int  bc$unit_init (IDB *idb, BC_UCB *ucb);

/* FDT routine for read functions */

    int  bc$read (IRP *irp, PCB *pcb, BC_UCB *ucb, CCB *ccb);

/* FDT routine for set mode and set characteristics functions */

    int  bc$setmode (IRP *irp, PCB *pcb, BC_UCB *ucb, CCB *ccb);

/* Start I/O routine */

    void bc$startio (IRP *irp, BC_UCB *ucb);

/* CSR mapping routine */

    int bc$csr_mapping (IDB *idb, DDB *ddb, CRB *crb);

/* Read bc635PCI current time routine */

    int bc$read_time (ADP *adp, BC_UCB *ucb);

/* Local routine that writes commands to the dual port RAM input area */

    int bc$send_bc_command (ADP *adp, BC_UCB *ucb, int cmd1, int cmd2);  

/* Interrupt service routine */

    void bc$interrupt (IDB *idb);

/* Driver fork routine entered when all I/O completed by interrupt service */

    void bc$iodone_fork (IRP *irp, void *not_used, BC_UCB *ucb);

/* Wait-for-interrupt timeout routine */

    void bc$wfi_timeout (IRP *irp, void *not_used, BC_UCB *ucb);

/* Standard breakpoint function prototype defintion */

    extern void ini$brk (void);

/* Pointer ot OPCOM mailbox unit control block */

    extern MB_UCB *sys$ar_oprmbx;


/*
 * DRIVER$INIT_TABLES - Initialize Driver Tables
 *
 * Functional description:
 *
 *   This routine completes the initialization of the DPT, DDT, and FDT
 *   structures.  If a driver image contains a routine named DRIVER$INIT_TABLES
 *   then this routine is called once by the $LOAD_DRIVER service immediately
 *   after the driver image is loaded or reloaded and before any validity checks
 *   are performed on the DPT, DDT, and FDT.  A prototype version of these
 *   structures is built into this image at link time from the
 *   VMS$VOLATILE_PRIVATE_INTERFACES.OLB library.  Note that the device related
 *   data structures (e.g. DDB, UCB, etc.) have not yet been created when this
 *   routine is called.  Thus the actions of this routine must be confined to
 *   the initialization of the DPT, DDT, and FDT structures which are contained
 *   in the driver image.
 *
 * Calling convention:
 *
 *   status = driver$init_tables ();
 *
 * Input parameters:
 *
 *   None.
 *
 * Output parameters:
 *
 *   None.
 *
 * Return value:
 *
 *   status     If the status is not successful, then the driver image will
 *              be unloaded.  Note that the ini_* macros used below will
 *              result in a return from this routine with an error status if
 *              an initialization error is detected.
 *
 * Implicit inputs:
 *
 *  driver$dpt, driver$ddt, driver$fdt
 *              These are the externally defined names for the prototype
 *              DPT, DDT, and FDT structures that are linked into this driver.
 *
 * Environment:
 * 
 *   Kernel mode, system context.
 */
 

int driver$init_tables ()  {

    /* Prototype driver DPT, DDT, and FDT will be pulled in from the
     * VMS$VOLATILE_PRIVATE_INTERFACES.OLB library at link time.
     */
    extern DPT driver$dpt;
    extern DDT driver$ddt;
    extern FDT driver$fdt;

    /* Finish initialization of the Driver Prologue Table (DPT) */

    ini_dpt_name        (&driver$dpt, "BCDRIVER");
    ini_dpt_adapt       (&driver$dpt, AT$_PCI);
    ini_dpt_defunits    (&driver$dpt, 1);
    ini_dpt_flags       (&driver$dpt, DPT$M_SMPMOD | DPT$M_SVP);
    ini_dpt_ucbsize     (&driver$dpt, sizeof(BC_UCB));
    ini_dpt_struc_init  (&driver$dpt, bc$struc_init );
    ini_dpt_struc_reinit(&driver$dpt, bc$struc_reinit );
    ini_dpt_ucb_crams   (&driver$dpt, NUMBER_CRAMS);
    ini_dpt_iohandles   (&driver$dpt, 2);
    ini_dpt_end         (&driver$dpt);

    /* Finish initialization of the Driver Dispatch Table (DDT) */

    ini_ddt_ctrlinit    (&driver$ddt, bc$ctrl_init);
    ini_ddt_unitinit    (&driver$ddt, bc$unit_init);
    ini_ddt_start       (&driver$ddt, bc$startio);
    ini_ddt_cancel      (&driver$ddt, ioc_std$cancelio);
    ini_ddt_csr_mapping (&driver$ddt, bc$csr_mapping);
    ini_ddt_end         (&driver$ddt);

    /* Finish initialization of the Function Decision Table (FDT)   */
    /*								    */
    /* The DIRECT indicates that this driver supports a 64-bit      */
    /* virtual address in the QIO P1 parameter for that function.   */
    /* This driver, therefore, supports 64-bit user buffers in all  */
    /* of its I/O functions.					    */

    ini_fdt_act (&driver$fdt, IO$_READLBLK,  exe_std$read,      DIRECT);
    ini_fdt_act (&driver$fdt, IO$_SETMODE,   bc$setmode,        DIRECT);
    ini_fdt_act (&driver$fdt, IO$_SETCHAR,   bc$setmode,        DIRECT);
    ini_fdt_act (&driver$fdt, IO$_SENSEMODE, exe_std$sensemode, DIRECT);
    ini_fdt_act (&driver$fdt, IO$_SENSECHAR, exe_std$sensemode, DIRECT);
    ini_fdt_end (&driver$fdt);

    /* If we got this far then everything worked, so return success. */

    return SS$_NORMAL;
}


/*
 * BC$STRUC_INIT - Device Data Structure Initialization Routine
 *
 * Functional description:
 *
 *   This routine is called once for each unit by the $LOAD_DRIVER service
 *   after that UCB is created.  At the point of this call the UCB has not
 *   yet been fully linked into the I/O database.  This routine is responsible
 *   for filling in driver specific fields that in the I/O database structures
 *   that are passed as parameters to this routine.
 *
 *   This routine is responsible for filling in the fields that are not
 *   affected by a RELOAD of the driver image.  In contrast, the structure
 *   reinitialization routine is responsible for filling in the fields that
 *   need to be corrected when (and if) this driver image is reloaded.
 *
 *   After this routine is called for a new unit, then the reinitialization
 *   routine is called as well.  Then the $LOAD_DRIVER service completes the
 *   integration of these device specific structures into the I/O database.
 *
 *   Note that this routine must confine its actions to filling in these I/O
 *   database structures and may not attempt to initialize the hardware device.
 *   Initialization of the hardware device is the responsibility of the 
 *   controller and unit initialization routines which are called some time
 *   later.
 *
 * Calling convention:
 *
 *   bc$struc_init (crb, ddb, idb, orb, ucb)
 *
 * Input parameters:
 *
 *   crb        Pointer to associated controller request block.
 *   ddb        Pointer to associated device data block.
 *   idb        Pointer to associated interrupt dispatch block.
 *   orb        Pointer to associated object rights block.
 *   ucb        Pointer to the unit control block that is to be initialized.
 *
 * Output parameters:   None.
 *
 * Return value:   None.
 *
 * Environment:
 * 
 *   Kernel mode, system context, IPL may be as high as 31 and may not be
 *   altered.
 *
 */ 

void bc$struc_init (CRB *crb, DDB *ddb, IDB *idb, ORB *orb, BC_UCB *ucb) {

    /* Initialize the fork lock and device IPL fields */

    ucb->ucb$r_ucb.ucb$b_flck = SPL$C_IOLOCK8;
    ucb->ucb$r_ucb.ucb$b_dipl = DEVICE_IPL;

    /* Device Characteristics are : Available (AVL), Real-time device (RTM),
     * Input device (IDV), Output device (ODV)
     */
    ucb->ucb$r_ucb.ucb$l_devchar = DEV$M_AVL | DEV$M_RTM | DEV$M_IDV | DEV$M_ODV;

    /* Set device is local, set device class, device type,
     *  and default buffer size.
     */
    ucb->ucb$r_ucb.ucb$l_devchar2  = DEV$M_LOC;
    ucb->ucb$r_ucb.ucb$b_devclass  = DC$_REALTIME;
    ucb->ucb$r_ucb.ucb$b_devtype   = DT$_USER9;
    ucb->ucb$r_ucb.ucb$w_devbufsiz = 128;

    /*
     * Set the initial state of the bc635PCI specific parameters in the UCB
     */
    ucb->ucb$l_bc_status = TFP_NO_ACTION; /* Don't do anything */
    ucb->ucb$l_bc_iohandle = 0;           /* No IO handles */
    ucb->ucb$l_bc_reghandle = 0;          /* No IO handles */
    ucb->ucb$l_bc_counter = 0;            /* Count of expected interrupts */
    ucb->ucb$l_bc_extra = 0;              /* Count of unexpected interrupts */

    return;
}


/*
 * BC$STRUC_REINIT - Device Data Structure Re-Initialization Routine
 *
 * Functional description:
 *
 *   This routine is called once for each unit by the $LOAD_DRIVER service
 *   immediately after the structure initialization routine is called.
 *
 *   Additionally, this routine is called once for each unit by the $LOAD_DRIVER
 *   service when a driver image is RELOADED.  Thus, this routine is
 *   responsible for filling in the fields in the I/O database structures
 *   that point into this driver image. Note that this routine must confine
 *   its actions to filling in these I/O database structures.
 *
 * Calling convention:
 *
 *   bc$struc_reinit (crb, ddb, idb, orb, ucb)
 *
 * Input parameters:
 *
 *   crb        Pointer to associated controller request block.
 *   ddb        Pointer to associated device data block.
 *   idb        Pointer to associated interrupt dispatch block.
 *   orb        Pointer to associated object rights block.
 *   ucb        Pointer to the unit control block that is to be initialized.
 *
 * Output parameters:   None.
 *
 * Return value:   None.
 *
 * Environment:
 * 
 *   Kernel mode, system context, IPL may be as high as 31 and may not be
 *   altered.
 *
 */ 

void bc$struc_reinit (CRB *crb, DDB *ddb, IDB *idb, ORB *orb, BC_UCB *ucb) {
    
    extern DDT driver$ddt;

    /* Setup the pointer from our DDB in the I/O database to the driver
     * dispatch table that's within this driver image.
     */
    ddb->ddb$ps_ddt = &driver$ddt;

    /* Setup the procedure descriptor and code entry addresses in the VEC
     * portion of the CRB in the I/O database to point to the interrupt
     * service routine that's within this driver image.
     */
    dpt_store_isr (crb, bc$interrupt);

    return;
}


/*
 * BC$READ_TIME - read time module
 *
 * Functional description:
 *
 *   This routine is module reads the current time from the device registers
 *   of the bc635PCI board. The time value is placed into the UCB structure.
 *
 * Calling convention:
 *
 *   status = bc$read_time (adp, ucb)
 *
 * Input parameters:
 *
 *   adp          Pointer to the adapter control block.
 *   ucb          Pointer to the unit control block.
 *
 * Output parameters:   None.
 *
 * Return value:
 *
 *   status     SS$_NORMAL      indicates that the unit was initialized 
 *                              successfully.
 *              SS$_BADPARM     indicates that an invalid parameter was passed
 *                              to ioc$map_io.
 * Environment:
 * 
 *   IPL 8, holds IOLOCK8 spinlock.
 */
 
int bc$read_time (ADP *adp, BC_UCB *ucb)
  {
    int     status;     /* Status to be returned to caller */

    /* Request the current time */
    ucb->ucb$l_bc_register = 1;
    status = ioc$write_io(adp,
                         &ucb->ucb$l_bc_reghandle,
                          BC_DR_TIMEREQ,
                          4,
                         &ucb->ucb$l_bc_register);

    /* Get the major time bits */
    status = ioc$read_io(adp,
                        &ucb->ucb$l_bc_reghandle,
                         BC_DR_TIME1,
                         4,
                        &ucb->ucb$l_bc_time1);

    /* Get the minor time bits */
    status = ioc$read_io(adp,
                        &ucb->ucb$l_bc_reghandle,
                         BC_DR_TIME0,
                         4,
                        &ucb->ucb$l_bc_time0);
    return status;
  }

/*
 * BC$CSR_MAPPING - CSR mapping module
 *
 * Functional description:
 *
 *   This routine is called prior to any controller or unit initialization.
 *   The function is to map the bc635PCI registers into space that can be
 *   accessed by the driver code when needed by way of the "iohandle". The
 *   Alpha VMS operating system takes care of all of the detail work to make
 *   it happen independent of the platform specifics. The IPL level is such
 *   that calls to ioc$map_io will work correctly.
 *
 *
 * Calling convention:
 *
 *   status = bc$csr_mapping (idb, ddb, crb)
 *
 * Input parameters:
 *
 *   idb        Pointer to associated interrupt dispatch block.
 *   ddb        Pointer to device data block.
 *   crb        Pointer to channel request block.
 *
 * Output parameters:
 *
 *   None.
 *
 * Return value:
 *
 *   status     SS$_NORMAL      indicates that the unit was initialized 
 *                              successfully.
 *              SS$_BADPARM     indicates that an invalid parameter was passed
 *                              to ioc$map_io.
 *              SS$_BADPARM     indicates that ioc$read_pci_config could not 
 *                              read the PCI configuration space address given.
 *              SS$_UNSUPPORTED indicates addressing mode give to ioc$map_io is
 *                              not supported on this system.
 *              SS$_INV_HW_ID   indicates that device does not have the bc635PCI
 *                              vendor or device ID as expected.
 *              SS$_IVADDR      indicates that the PCI configuration could not
 *                              be mapped, the configureation registers could
 *                              not be read, or the remaining hardware registers
 *                              could not mapped and read.
 *
 * Environment:
 * 
 *   IPL 8, holds IOLOCK8 spinlock.
 */
 
int bc$csr_mapping (IDB *idb, DDB *ddb, CRB *crb)  
  {

    ADP    *adp;
    BC_UCB *ucb;         /* The UCB for the IRIG-B board */
    MCJ    *mcj;         /* The magic cookie jar for IO handles */
    int     results;     /* Status to be returned to caller */
    int     config_1;    /* Place to hold data from PCI configuration area */
    int     config_2;    /* Place to hold data from PCI configuration area */
    uint64  pci_address; /* Place to hold 64 bit PCI address */


    /*
     * Get the ADP, UCB, and the magic cookie jar addresses
     */
    adp = idb->idb$ps_adp;
    ucb = (BC_UCB *)ddb->ddb$l_ucb;
    mcj = (MCJ *)idb->idb$q_csr;

    /*
     * Read the vendor and device configuration parameters and
     * verify that this is the correct device.
     */
    results = ioc$read_pci_config(adp, 
                                  crb->crb$l_node, 
                                  PCI$K_VENDOR_ID,
                                  2,
                                 &config_1);
    if (results == SS$_NORMAL)
    {
      /* Get the vendor ID bits and confirm correct board */
      config_1 = config_1 & 0x0000FFFF;
      if (config_1 != 0x12E2)
      {
        /* Incorrect vendor id, force a fail (invalid media address) */
        results = SS$_IVADDR;
      };
    };
    if (results == SS$_NORMAL)
    {
      results = ioc$read_pci_config(adp, 
                                    crb->crb$l_node,
                                    PCI$K_DEVICE_ID,
                                    2,
                                   &config_2);
      if (results == SS$_NORMAL)
      {
        /* Get the device ID bits and confirm correct board */
        config_2 = (config_2 >> 16) & 0x0000FFFF;
        if (config_2 != 0x4013)
        {
          /* Incorrect device id, force a fail (invalid i/o channel) */
          results = SS$_IVCHAN;
        };
      };
    };

    /*
     * Read bc635PCI register offset, map the offset, and save iohandle
     */
    if (results == SS$_NORMAL)
    {
      /* Read the base address 0 of the bc635PCI */
      results = ioc$read_pci_config(adp,
                                    crb->crb$l_node,
                                    PCI$K_BASE_ADDRESS_0,
                                    4,
                                   &config_1);
      if (results == SS$_NORMAL)
      {
        /* Map base address to an reghandle for access to the bc635PCI */
        pci_address = config_1 & PCI$M_BASE_ADDRESS_BITS_31_4;
        results = ioc$map_io(adp, 
                             crb->crb$l_node, 
                            &pci_address, 
                             0x0040,
                             IOC$K_BUS_MEM_BYTE_GRAN, 
                            &ucb->ucb$l_bc_reghandle);
        /* Save off the handle for the bc635PCI */
        mcj->mcj$q_cookies[0] = ucb->ucb$l_bc_reghandle;
      };
    };

    /*
     * Read bc635PCI DPRAM offset, map the offset, and save iohandle
     */
    if (results == SS$_NORMAL)
    {
      /* Read the base address 1 of the bc635PCI */
      results = ioc$read_pci_config(adp,
                                    crb->crb$l_node,
                                    PCI$K_BASE_ADDRESS_1,
                                    4,
                                   &config_1);
      if (results == SS$_NORMAL)
      {
        /* Map base address to an iohandle for access to the bc635PCI */
        pci_address = config_1 & PCI$M_BASE_ADDRESS_BITS_31_4;
        results = ioc$map_io(adp, 
                             crb->crb$l_node, 
                            &pci_address, 
                             0x1000,
                             IOC$K_BUS_MEM_BYTE_GRAN, 
                            &ucb->ucb$l_bc_iohandle);
        /* Save off the handle for the bc635PCI */
        mcj->mcj$q_cookies[1] = ucb->ucb$l_bc_iohandle;
      };
    };

    return results;
  }

/*
 * BC$CTRL_INIT - Controller Initialization Routine
 *
 * Functional description:
 *
 *   This routine is called once for each controller by the $LOAD_DRIVER
 *   service after a new unit control block has been created, initialized,
 *   and fully integrated into the I/O database.
 *
 *   This routine is also called for each controller during power fail 
 *   recovery.
 *
 * Calling convention:
 *
 *   status = bc$ctrl_init (idb, ddb, crb)
 *
 * Input parameters:
 *
 *   idb        Pointer to associated interrupt dispatch block.
 *   ddb        Pointer to device data block.
 *   crb        Pointer to channel request block.
 *
 * Output parameters:
 *
 *   None.
 *
 * Return value:
 *
 *   status     SS$_NORMAL      indicates that the unit was initialized 
 *                              successfully.
 *
 * Environment:
 * 
 *   Kernel mode, system context, IPL 31
 */
 
int bc$ctrl_init (IDB *idb, DDB *ddb, CRB *crb)  
  {
    int     results; /* Status to be returned to caller */

    /* No special actions needed for PCI bus controller */
    /* Always return a good status                      */
    results = SS$_NORMAL;

    return results;
  }

/*
 * BC$UNIT_INIT - Unit Initialization Routine
 *
 * Functional description:
 *
 *   This routine is called once for each unit by the $LOAD_DRIVER service
 *   after a new unit control block has been created, initialized, and
 *   fully integrated into the I/O database.
 *
 *   This routine is also called for each unit during power fail recovery.
 *
 *   It is the responsibility of this routine to bring unit "on line" and
 *   to make it ready to accept I/O requests.
 *
 * Calling convention:
 *
 *   status = bc$unit_init (idb, ucb)
 *
 * Input parameters:
 *
 *   idb        Pointer to associated interrupt dispatch block.
 *   ucb        Pointer to the unit control block that is to be initialized.
 *
 * Output parameters:
 *
 *   None.
 *
 * Return value:
 *
 *   status     SS$_NORMAL indicates that the unit was initialized successfully.
 *              SS$_IVADDR indicates that an unexpected ISA I/O address or IRQ
 *                         level was detected.
 *
 * Environment:
 * 
 *   Kernel mode, system context, IPL 31.
 */
 
int bc$unit_init (IDB *idb, BC_UCB *ucb)  
{

    ADP *adp;
    int tfp_data_1;            /* Data from or for bc635PCI board */
    int tfp_data_2;            /* Data from or for bc635PCI board */
    int status;
    int loop_counter;
    int lockup;
    int offset;


    /* Set device initially offline (for error exits) and initialize other
     * UCB cells.
     */
    ucb->ucb$r_ucb.ucb$v_online = 0;

    /* This driver can service only a single unit per DDB and IDB.  Thus,
     * make the single unit the permanent owner of the IDB.  This facilitates
     * getting the UCB address in our interrupt service routine.
     */
    idb->idb$ps_owner = &(ucb->ucb$r_ucb); 

    /*
     * Get the pointer to the ADP
     */
    adp = ucb->ucb$r_ucb.ucb$ps_adp;            /* Pointer to our ADP */

    /*
     * Read the offset to the input area of Dual Port RAM
     */
    status = ioc$read_io(adp,
                        &ucb->ucb$l_bc_iohandle,
                         BC_TOP,
                         1,
                        &ucb->ucb$l_bc_register);

    /* Check status and continue only if things worked up to now */
    if (status == SS$_NORMAL)
    {
      offset = ucb->ucb$l_bc_register & 0x000000FF;
      status = ioc$read_io(adp,
                          &ucb->ucb$l_bc_iohandle,
                           BC_TOP_1,
                           1,
                          &ucb->ucb$l_bc_register);

      /* Check status and continue only if things worked up to now */
      if (status == SS$_NORMAL)
      {
        offset = offset | (ucb->ucb$l_bc_register & 0x0000FF00);
        ucb->ucb$l_bc_put_offset = offset;
      };
    };

    /*
     * Read the offset to the output area of Dual Port RAM
     */
    if (status == SS$_NORMAL)
    {
      status = ioc$read_io(adp,
                          &ucb->ucb$l_bc_iohandle,
                           BC_TOP_2,
                           1,
                          &ucb->ucb$l_bc_register);

      /* Check status and continue only if things worked up to now */
      if (status == SS$_NORMAL)
      {
        offset = ucb->ucb$l_bc_register & 0x000000FF;
        status = ioc$read_io(adp,
                            &ucb->ucb$l_bc_iohandle,
                             BC_TOP_3,
                             1,
                            &ucb->ucb$l_bc_register);

        /* Check status and continue only if things worked up to now */
        if (status == SS$_NORMAL)
        {
          offset = offset | (ucb->ucb$l_bc_register & 0x0000FF00);
          ucb->ucb$l_bc_get_offset = offset;
        };
      };
    };

    /*
     * Read the offset to the GSP packet area of Dual Port RAM
     */
    if (status == SS$_NORMAL)
    {
      status = ioc$read_io(adp,
                          &ucb->ucb$l_bc_iohandle,
                           BC_TOP_4,
                           1,
                          &ucb->ucb$l_bc_register);

      /* Check status and continue only if things worked up to now */
      if (status == SS$_NORMAL)
      {
        offset = ucb->ucb$l_bc_register & 0x000000FF;
        status = ioc$read_io(adp,
                            &ucb->ucb$l_bc_iohandle,
                             BC_TOP_5,
                             1,
                            &ucb->ucb$l_bc_register);

        /* Check status and continue only if things worked up to now */
        if (status == SS$_NORMAL)
        {
          offset = offset | (ucb->ucb$l_bc_register & 0x0000FF00);
          ucb->ucb$l_bc_gps_offset = offset;
        };
      };
    };

    /*
     * Read the offset to the year area of Dual Port RAM
     */
    if (status == SS$_NORMAL)
    {
      status = ioc$read_io(adp,
                          &ucb->ucb$l_bc_iohandle,
                           BC_TOP_6,
                           1,
                          &ucb->ucb$l_bc_register);

      /* Check status and continue only if things worked up to now */
      if (status == SS$_NORMAL)
      {
        offset = ucb->ucb$l_bc_register & 0x000000FF;
        status = ioc$read_io(adp,
                            &ucb->ucb$l_bc_iohandle,
                             BC_TOP_7,
                             1,
                            &ucb->ucb$l_bc_register);

        /* Check status and continue only if things worked up to now */
        if (status == SS$_NORMAL)
        {
          offset = offset | (ucb->ucb$l_bc_register & 0x0000FF00);
          ucb->ucb$l_bc_year_offset = offset;
        };
      };
    };

    /*
     * Read and ignore the time from the bc635PCI board three times.
     *   (because datum sample driver code for DEC UNIX did this)
     */
    status = bc$read_time(adp, ucb);
    status = bc$read_time(adp, ucb);
    status = bc$read_time(adp, ucb);

/*
 * Set the clock source to "I" - internal 10 MHz VCXO
 */

/*
 *   if (status == SS$_NORMAL)
 *   {
 *     tfp_data_1 = 0x00004920;
 *     tfp_data_2 = 0x00000000;
 *     status = bc$send_bc_command (adp, ucb, tfp_data_1, tfp_data_2);
 *   };
 */

/*
 * Set the time code format to "B" time code - IRIG B
 */

/*
 *   if (status == SS$_NORMAL)
 *   {
 *     tfp_data_1 = 0x00004215;
 *     tfp_data_2 = 0x00000000;
 *     status = bc$send_bc_command (adp, ucb, tfp_data_1, tfp_data_2);
 *   };
 */

/*
 * Set the time code modulation type to "M" - Amplitude modulation
 */

/*
 *   if (status == SS$_NORMAL)
 *   {
 *     tfp_data_1 = 0x00004D16;
 *     tfp_data_2 = 0x00000000;
 *     status = bc$send_bc_command (adp, ucb, tfp_data_1, tfp_data_2);
 *   };
 */

/*
 * Set a timing mode of "0" time code - IRIG B if everything is working
 */
    if (status == SS$_NORMAL)
    {
      tfp_data_1 = 0x00003010;
      tfp_data_2 = 0x00000000;
      status = bc$send_bc_command (adp, ucb, tfp_data_1, tfp_data_2);
    };

    /* Continue only if prior operations returned good status */
    if (status == SS$_NORMAL)
    {
      /* Enable PCI bus interrupts from the bc635PCI card slot */
      status = ioc$node_function(ucb->ucb$r_ucb.ucb$l_crb,
                                 IOC$K_ENABLE_INTR);

      /* Clear the 1 PPS interrupt bit in the bc635PCI registers */
      ucb->ucb$l_bc_register = TFP_I_SECOND;
      status = ioc$write_io(adp,
                           &ucb->ucb$l_bc_reghandle,
                            BC_DR_INTSTAT,
                            4,
                           &ucb->ucb$l_bc_register);

      /* Enable the 1 PPS interrupt in the bc635PCI registers */
      ucb->ucb$l_bc_register = TFP_I_SECOND;
      status = ioc$write_io(adp,
                           &ucb->ucb$l_bc_reghandle,
                            BC_DR_MASK,
                            4,
                           &ucb->ucb$l_bc_register);

    };

    /* Continue only if prior operations returned good status */
    if (status == SS$_NORMAL)
    {
      /* Mark the device as "on line" and ready to accept I/O requests */
      ucb->ucb$r_ucb.ucb$v_online = 1;
    };

    return status;
}

/*
 * BC$SEND_BC_COMMAND - Send a command to the bc635PCI by DPRAM Routine
 *
 * Functional description:
 *
 *  This module writes the command to the dual ported ram using the bc635PCI
 *  hand shake based upon the ACK register bits.
 *
 *  The length sent is always 8 bytes.
 *
 * Calling convention:
 *
 *   status = bc$send_bc_command (adp, ucb, cmd1, cmd2)
 *
 * Input parameters:
 *
 *   adp          Pointer to the adapter control block.
 *   ucb          Pointer to the unit control block.
 *   cmd1         First 4 bytes of the command.
 *   cmd2         Second 4 bytes of the command.
 *
 * Output parameters:
 *
 *   None.
 *
 * Return value:
 *
 *   status     SS$_NORMAL       indicates that the command was sent to the
 *                               bc635PCI board successfully.
 *              SS$_CTRLERR      indicates that the bc635PCI board did not
 *                               respond to the command placed into the dual
 *                               port RAM.
 *              SS$_BADPARAM     indicates that one or more of the parameters
 *                               invalid.
 *              SS$_UNSUPPORTED  indicates that the transaction length is not
 *                               supported on this bus or platform.
 *
 * Environment:
 * 
 *   Kernel mode, system context, IPL 31.
 */
 
int bc$send_bc_command (ADP *adp, BC_UCB *ucb, int cmd1, int cmd2)

{

    int tfp_data;            /* Data from or for bc635PCI board */
    int status;
    int loop_counter;
    int lockup;
    int data_words[8];
    int offset;
    int work;


    /* Build data words to send that look like the words received. */
    work = cmd1 & 0x000000FF;
    data_words[0] = work | (work << 8) | (work << 16) | (work << 24);
    work = (cmd1 >> 8) & 0x000000FF;
    data_words[1] = work | (work << 8) | (work << 16) | (work << 24);
    work = (cmd1 >> 16) & 0x000000FF;
    data_words[2] = work | (work << 8) | (work << 16) | (work << 24);
    work = (cmd1 >> 24) & 0x000000FF;
    data_words[3] = work | (work << 8) | (work << 16) | (work << 24);
    work = cmd2 & 0x000000FF;
    data_words[4] = work | (work << 8) | (work << 16) | (work << 24);
    work = (cmd2 >> 8) & 0x000000FF;
    data_words[5] = work | (work << 8) | (work << 16) | (work << 24);
    work = (cmd2 >> 16) & 0x000000FF;
    data_words[6] = work | (work << 8) | (work << 16) | (work << 24);
    work = (cmd2 >> 24) & 0x000000FF;
    data_words[7] = work | (work << 8) | (work << 16) | (work << 24);

    /* Write the command to the dual ported ram */
    offset = ucb->ucb$l_bc_put_offset;
    offset = offset + BC_ROM_SIZE;

    /* Write the data words to the input area of the dual ported ram */
    for (work = 0; work < 8; work++)
    {
      ucb->ucb$l_bc_register = data_words[work];
      status = ioc$write_io(adp, 
                           &ucb->ucb$l_bc_iohandle,
                            offset,
                            1,
                           &ucb->ucb$l_bc_register);

      /* Adjust the offset*/
      offset = offset + 1;
    };

    /* Continue only if the prior operations worked */
    if (status == SS$_NORMAL)
    {
      /* Clear ACK register command received bit */
      ucb->ucb$l_bc_register = TFP_A_CMDIN;
      status = ioc$write_io(adp,
                           &ucb->ucb$l_bc_reghandle,
                            BC_DR_ACK,
                            4,
                           &ucb->ucb$l_bc_register);

          /* Get the ACK register to verify that it is cleared */
          status = ioc$read_io(adp,
                              &ucb->ucb$l_bc_reghandle,
                               BC_DR_ACK,
                               4,
                              &ucb->ucb$l_bc_register);
          tfp_data = ucb->ucb$l_bc_register;

      /* Continue only if the prior operations worked */
      if (status == SS$_NORMAL)
      {
        /* Set the new command ready bit in the ACK register */
        ucb->ucb$l_bc_register = TFP_A_CMDRDY;
        status = ioc$write_io(adp,
                             &ucb->ucb$l_bc_reghandle,
                              BC_DR_ACK,
                              4,
                             &ucb->ucb$l_bc_register);

        /* Continue only if the prior operations worked */
        if (status == SS$_NORMAL)
        {
          /* Get the ACK register command received bit */
          status = ioc$read_io(adp,
                              &ucb->ucb$l_bc_reghandle,
                               BC_DR_ACK,
                               4,
                              &ucb->ucb$l_bc_register);

          /* Loop waiting for the command received bit to be set or timeout */
          lockup = 0x0000FFFF;
          loop_counter = 0;
          tfp_data = ucb->ucb$l_bc_register & TFP_A_CMDIN;
          while ((loop_counter < lockup) && 
                 (tfp_data != TFP_A_CMDIN) &&
                 (status == SS$_NORMAL))
          {
            /* Count the pass through the loop */
            loop_counter = loop_counter + 1;

            /* Get the ACK register command received bit */
            status = ioc$read_io(adp,
                                &ucb->ucb$l_bc_reghandle,
                                 BC_DR_ACK,
                                 4,
                                &ucb->ucb$l_bc_register);
            tfp_data = ucb->ucb$l_bc_register & TFP_A_CMDIN;
          };

          /* If the command was an TFP data request, read back the response */
          if ((cmd1 & 0x000000FF) == 0x19)
          {
            offset = ucb->ucb$l_bc_get_offset + BC_ROM_SIZE;
            for (work = 0; work < 24; work++)
            {
              status = ioc$read_io(adp, 
                                  &ucb->ucb$l_bc_iohandle,
                                   offset,
                                   1,
                                  &ucb->ucb$l_bc_register);
              ucb->ucb$l_bc_response[work] = 
                                 (uint8) (ucb->ucb$l_bc_register & 0x000000FF);

              /* Adjust the offset*/
              offset = offset + 1;
            };
          };

          /* Check for a failure to handshake with the bc635 board */
          if ((loop_counter >= lockup) && 
              (tfp_data != TFP_A_CMDIN))
          {
            /* Return error status because ACK bit was never set */
            status = SS$_CTRLERR;
          };

        };

      };

    };

    return status;
}

/*
 * BC$SETMODE - FDT Routine for Set Mode and Set Characteristics
 *
 * Functional description:
 *
 *   This routine is called by the FDT dispatcher in the $QIO system service
 *   to process set mode and set characteristics functions.  This FDT routine
 *   completes the I/O request without sending it to the driver start I/O
 *   routine. The user buffer address is contained in irp$q_qio_p1 ($QIO 
 *   P1 parameter) on input and will be treated as a 64-bit address.
 *
 *   This routine places the year and offset value into the UCB. The TFP_LOAD 
 *   flag is used to signal the interrupt routine to write this information to
 *   the bc635PCI card when servicing the next interrupt.
 *
 *   Since this is an upper-level FDT routine, this routine always returns
 *   the SS$_FDT_COMPL status.  The $QIO status that is to be returned to
 *   the caller of the $QIO system service is returned indirectly by the
 *   FDT completion routines (e. g. exe_std$abortio, exe_std$finishio) via
 *   the FDT context structure.
 *
 * Calling convention:
 *
 *   status = bc$setmode (irp, pcb, ucb, ccb)
 *
 * Input parameters:
 *
 *   irp        Pointer to I/O request packet
 *   pcb        Pointer process control block
 *   ucb        Pointer to unit control block
 *   ccb        Pointer to channel control block
 *
 * Output parameters:
 *
 *   None.
 *
 * Return value:
 *
 *   status     SS$_FDT_COMPL
 *
 * Environment:
 * 
 *   Kernel mode, user process context, IPL 2.
 */

int  bc$setmode (IRP *irp, PCB *pcb, BC_UCB *ucb, CCB *ccb) {

    /* Define a structure that corresponds to the layout of the caller's 
     * set mode or set characteristics buffer and declare a local pointer
     * to a structure of this type.
     */
    typedef struct { 
                    int irig_year;
                    int irig_offset; 
                   } SETMODE_BUF;

    #pragma __required_pointer_size __save
    #pragma __required_pointer_size __long

    /* Define a type for a 64-bit pointer to a SETMODE_BUF structure.
     */
    typedef SETMODE_BUF *SETMODE_BUF_PQ;

    #pragma __required_pointer_size __restore

    /* This must be a pointer to a 64-bit address since it will be containing 
     *  the address of a user buffer which may be a 64-bit or a 32-bit value.  
     */
    SETMODE_BUF_PQ setmode_bufp;		

    /* The caller passes the address of their setmode buffer in the $QIO P1
     * parameter.
     */
    setmode_bufp = (SETMODE_BUF_PQ) irp->irp$q_qio_p1;

    /* Assure that the caller's setmode buffer is readable by the caller.
     * If not, abort the I/O request now with an ACCVIO status and return
     * back to the FDT dispatcher in the $QIO system service.
     */
    if (! ( __PAL_PROBER (setmode_bufp, sizeof(SETMODE_BUF)-1, irp->irp$b_rmod) ))
    {
        return ( call_abortio (irp, pcb, (UCB *)ucb, SS$_ACCVIO) ); 
    };

    /* If function is SETCHAR then reject as unsupported */
    if (irp->irp$v_fcode == IO$_SETCHAR)
    {
        return ( call_abortio (irp, pcb, (UCB *)ucb, SS$_UNSUPPORTED) ); 
    };

    /* Set the year and offset into the ucp, set flag to have them updated */
    ucb->ucb$l_bc_year   = setmode_bufp->irig_year;
    ucb->ucb$l_bc_offset = setmode_bufp->irig_offset;
    ucb->ucb$l_bc_status = TFP_LOAD;

    /* Finish the IO; return SS$_FDT_COMPL to the FDT dispatcher in the $QIO
     * system service.
     */
    return ( call_finishio (irp, (UCB *)ucb, SS$_NORMAL, 0) );
}

/*
 * BC$READ - FDT Routine for Read Function Codes 
 *
 * Functional description:
 *
 *   This routine is called by the FDT dispatcher in the $QIO system service
 *   to process read functions.  This FDT routine validates the request,
 *   allocates a buffered I/O packet, formats and copies the contents of the
 *   user buffer into the buffered I/O packet, and queues the IRP to this
 *   driver's start I/O routine.  The user buffer address is contained in
 *   irp$q_qio_p1 ($QIO P1 parameter) on input and will be treated as a 
 *   64-bit address.              
 *                                
 *   Since this is an upper-level FDT routine, this routine always returns
 *   the SS$_FDT_COMPL status.  The $QIO status that is to be returned to
 *   the caller of the $QIO system service is returned indirectly by the
 *   FDT completion routines (e. g. exe_std$abortio, exe_std$qiodrvpkt) via
 *   the FDT context structure.
 *
 * Calling convention:
 *
 *   status = bc$read (irp, pcb, ucb, ccb)
 *
 * Input parameters:
 *
 *   irp        Pointer to I/O request packet
 *   pcb        Pointer process control block
 *   ucb        Pointer to unit control block
 *   ccb        Pointer to channel control block
 *
 * Output parameters:
 *
 *   None.
 *
 * Return value:
 *
 *   status     SS$_FDT_COMPL
 *
 * Environment:
 * 
 *   Kernel mode, user process context, IPL 2.
 */

int bc$read (IRP *irp, PCB *pcb, BC_UCB *ucb, CCB *ccb) {

    int status;


    status = SS$_FDT_COMPL;

    return ( status );

}

/*
 * BC$STARTIO - Start I/O Routine
 *
 * Functional description:
 *
 *   This routine is the driver start I/O routine.  This routine is called
 *   by ioc_std$initiate to process the next I/O request that has been
 *   queued to this device.  For this driver, the only function that is
 *   passed to the start I/O routine is a write operation.
 *
 *   Before this routine is called, ucb$v_cancel, ucb$v_int, ucb$v_tim, and
 *   ucb$v_timout are cleared. The ucb$l_svapte, ucb$l_boff, and ucb$l_bcnt
 *   cells are set in  ioc_std$initiate from their corresponding IRP cells. 
 *   Unlike their IRP counterparts, these UCB cells are working storage and 
 *   can be changed by a driver. 
 *
 *   This routine acquires the device lock and raises IPL to device IPL.
 *   The any pending data writes (from the setmode command) are performed.
 *   The interrupt expected flag is set. And then interrupts are enabled.
 *   The device lock is restored and the original IPL is restored via wfikpch
 *   before this routine returns to its caller.
 *
 * Calling convention:
 *
 *   bc$startio (irp, ucb)
 *
 * Input parameters:
 *
 *   irp        Pointer to I/O request packet
 *   ucb        Pointer to unit control block
 *
 * Output parameters:
 *
 *   None.
 *
 * Return value:
 *
 *   None.
 *
 * Environment:
 * 
 *   Kernel mode, system context, fork IPL, fork lock held.
 */

void bc$startio (IRP *irp, BC_UCB *ucb) {

    int  orig_ipl;
    int  command_part_1;
    int  command_part_2;
    int  command_status;
    ADP *adp;

    /* Acquire the device lock, raise IPL, saving original IPL */
    device_lock (ucb->ucb$r_ucb.ucb$l_dlck, RAISE_IPL, &orig_ipl);

    /* Get the address of the ADP */
    adp = ucb->ucb$r_ucb.ucb$ps_adp;

    /* Clear the cancel and timeout indicators in the UCB */
    ucb->ucb$r_ucb.ucb$v_cancel = FALSE;
    ucb->ucb$r_ucb.ucb$v_timout = FALSE;

    /* Handle any writes to the device of information from SETMODE commands */
    if (ucb->ucb$l_bc_status == TFP_LOAD)
    {

      /* Build the command to set the YEAR */
      command_part_1 = 0x00000013;
      command_part_1 = command_part_1 | 
                       (((ucb->ucb$l_bc_year >> 8) & 0x00FF) << 8);
      command_part_1 = command_part_1 | 
                       (((ucb->ucb$l_bc_year) & 0x00FF) << 16);
      command_part_2 = 0x00000000;

      /* Send the YEAR command */
      command_status = bc$send_bc_command(adp, 
                                          ucb, 
                                          command_part_1,
                                          command_part_2);

      /* Read back the year from the bc635PCI board */
/*
 *     command_part_1 = 0x00001319;
 *     command_part_2 = 0x00000000;
 *     command_status = bc$send_bc_command(adp, 
 *                                         ucb, 
 *                                         command_part_1,
 *                                         command_part_2);
 */

      /* Save off the year into the ucb */
/*
 *     ucb->ucb$l_bc_sync = (int)ucb->ucb$l_bc_response[0];
 *     ucb->ucb$l_bc_n1   = 
 *        (int)((ucb->ucb$l_bc_response[2] << 8) | ucb->ucb$l_bc_response[1]);
 */

      /* Send any non-zero OFFSET command if the YEAR command was sent OK */
      if ((command_status == SS$_NORMAL) && 
          (ucb->ucb$l_bc_offset != 0))
      {
        /* Add the turn off JAM-sync command before the offset command */
        command_part_1 = 0x00000021;
        command_part_2 = 0x00000000;
 
        /* Send the JAM-SYNC off command */
        command_status = bc$send_bc_command(adp, 
                                            ucb, 
                                            command_part_1,
                                            command_part_2);

        /* Continue only if first part worked */
        if (command_status == SS$_NORMAL)
        {
          /* Insert the command */
          command_part_1 = 0x00000017;

          /* Add the first 8 bits of the offset */
          command_part_1 = command_part_1 |
                           (0x0000FF00 & (ucb->ucb$l_bc_offset >> 16));

          /* Add the second 8 bits of the offset */
          command_part_1 = command_part_1 |
                           (0x00FF0000 & ucb->ucb$l_bc_offset);

          /* Add the third 8 bits of the offset */
          command_part_1 = command_part_1 |
                           (0xFF000000 & (ucb->ucb$l_bc_offset << 16));

          /* Add the last 8 bits of the offset */
          command_part_2 = 0x00FF & ucb->ucb$l_bc_offset;

          /* Send the OFFSET command */
          command_status = bc$send_bc_command(adp, 
                                              ucb, 
                                              command_part_1,
                                              command_part_2);
        };

      };

      /* Abort the IO if unable to send the commands */
      if (command_status != SS$_NORMAL)
      {
        /* Cancel the IO operation */
        ucb->ucb$r_ucb.ucb$v_cancel = TRUE;
      };

    };

    /* Show that we are expecting to read a new time value */
    ucb->ucb$l_bc_status = TFP_READ;
    ucb->ucb$r_ucb.ucb$v_int = TRUE;

    /* Set up a wait for the completion of the I/O by using the wfikpch macro.
     * Wfikpch will restore the device lock and restore IPL.  When the next
     * interrupt from the bc635PCI board occurs, the bc$interrupt routine will
     * queue the bc$iodone_fork routine.  If the I/O does not complete within
     * BC_WFI_TMO seconds, then exe$timeout will call bc$wfi_timeout.
     */
    wfikpch (bc$iodone_fork, bc$wfi_timeout, irp, 0, (UCB *)ucb, BC_WFI_TMO,
             orig_ipl);

    return;
}



/*
 * BC$INTERRUPT - Interrupt Service Routine
 *
 * Functional description:
 *
 *   This is the interrupt service routine for the bc635PCI board. This routine
 *   is called by the system interrupt dispatcher.
 *
 *   This routine will read the information from the bc635PCI board and place
 *   the information into the UCB.
 *
 *   If the interrupt is not expected by an active I/O on this device then
 *   it is simply dismissed.
 *
 *
 * Calling convention:
 *
 *   bc$interrupt (idb)
 *
 * Input parameters:
 *
 *   idb        Pointer to interrupt dispatch block
 *
 * Output parameters:
 *
 *   None.
 *
 * Return value:
 *
 *   None.
 *
 * Environment:
 * 
 *   Kernel mode, system context, device IPL.
 *
 */

void bc$interrupt (IDB *idb)  {

    BC_UCB *ucb;
    ADP    *adp;
    int     status;
    int     offset;

    /*
     * Get the UCB from the IDB owner field which was set up by the bc$unit_init
     * routine.
     *
     */
    ucb = (BC_UCB *) idb->idb$ps_owner;

    /* Acquire the device lock.  We are already at device IPL */
    device_lock (ucb->ucb$r_ucb.ucb$l_dlck, NORAISE_IPL, NOSAVE_IPL);

    /* Get the address of the ADP */
    adp = ucb->ucb$r_ucb.ucb$ps_adp;

    /* Read the time from the bc635PCI card. */
    status = bc$read_time(adp, ucb);

    /* Clear the 1 PPS interrupt bit */
    ucb->ucb$l_bc_register = TFP_I_SECOND;
    status = ioc$write_io(adp,
                         &ucb->ucb$l_bc_reghandle,
                          BC_DR_INTSTAT,
                          4,
                         &ucb->ucb$l_bc_register);

    /* Enable the 1 PPS interrupt */
    ucb->ucb$l_bc_register = TFP_I_SECOND;
    status = ioc$write_io(adp,
                         &ucb->ucb$l_bc_reghandle,
                          BC_DR_MASK,
                          4,
                         &ucb->ucb$l_bc_register);

    /* If interrupt is expected, then process it, otherwise ignore it */
    if (ucb->ucb$r_ucb.ucb$v_int) 
    {
      /* Then count the interrupt */
      ucb->ucb$l_bc_counter = ucb->ucb$l_bc_counter + 1;

/*
 * NOTE: Any additional information that needs to be read back from the IRIG
 *       board should be read back at this point.
 */

      /* Then complete the interrupt processing */
      ucb->ucb$r_ucb.ucb$v_int = 0;
      ucb->ucb$r_ucb.ucb$v_tim = 0;
      exe_std$queue_fork( (FKB *)ucb );
    }
    else
    {
      /* Just count the unexpected interrupt */
      ucb->ucb$l_bc_extra = ucb->ucb$l_bc_extra + 1;
    };

    /* Restore the device lock, stay at device IPL */
    device_unlock (ucb->ucb$r_ucb.ucb$l_dlck, NOLOWER_IPL, SMP_RESTORE);

    /* return back to interrupt dispatcher */
    return;
}


/*
 * BC$IODONE_FORK - I/O Completion Fork Routine
 *
 * Functional description:
 *
 *   This is the fork routine which passes the current I/O request on to
 *   I/O postprocessing.  This routine is queued by the interrupt service
 *   routine when the I/O request has been completed.
 *
 * Calling convention:
 *
 *   bc$iodone_fork (irp, not_used, ucb)
 *
 * Input parameters:
 *
 *   irp        Pointer to I/O request packet
 *   not_used   Unused fork routine parameter fr4
 *   ucb        Pointer to unit control block
 *
 * Output parameters:
 *
 *   None.
 *
 * Return value:
 *
 *   None.
 *
 * Environment:
 * 
 *   Kernel mode, system context, fork IPL, fork lock held.
 */

void bc$iodone_fork (IRP *irp, void *not_used, BC_UCB *ucb) {

    ADP    *adp;
    uint64  time_value;
    uint64  time_value_2;
    int     status;
    struct RAW_TIME_STRUCT time_data;
    int     time_data_size;
    void    *next_byte;
    int     offset;
    int     tfp_data_1;
    int     tfp_data_2;


    /*
     * If the request was cancelled, set an abort status
     *
     * If the request was timed out, set an timeout status
     *
     * otherwise return successful status
     *
     */
    if (ucb->ucb$r_ucb.ucb$v_cancel == TRUE) 
    {
        status = SS$_ABORT;
    } 
    else 
    {
      if (ucb->ucb$r_ucb.ucb$v_timout == TRUE) 
      {
        status = SS$_TIMEOUT;

/*****************************************************************************/
/*****************************************************************************/

      /*
       * Get the pointer to the ADP address
       */
      adp = ucb->ucb$r_ucb.ucb$ps_adp;            /* Pointer to our ADP */

        /* Read the time */
        status = bc$read_time(adp, ucb);

        /* For testing, set to return a normal status */
        status = SS$_NORMAL;

        /* Create the 64 bit time data information */
        time_value   = ucb->ucb$l_bc_time0;
        time_value   = time_value << 32;
        time_value_2 = ucb->ucb$l_bc_time1;
        time_value   = time_value | time_value_2;

        /* Setup offset to the year value stored in Dual Port RAM */
        offset = ucb->ucb$l_bc_year_offset;
        offset = offset + BC_ROM_SIZE;

        /* Read the first byte of the year from the Dual Port ram */
        status = ioc$write_io(adp, 
                             &ucb->ucb$l_bc_iohandle,
                              offset,
                              1,
                             &ucb->ucb$l_bc_register);

        /* Save off the high order bits */
        ucb->ucb$l_bc_year = 0xFF00 & (ucb->ucb$l_bc_register << 8);

        /* Adjust the offset */
        offset = offset + 1;

        /* If still OK, read the second byte of the year from the Dual Port RAM */
        if (status == SS$_NORMAL)
        {
          status = ioc$write_io(adp,
                               &ucb->ucb$l_bc_iohandle,
                                offset,
                                1,
                               &ucb->ucb$l_bc_register);

          /* Save off the low order bits */
          ucb->ucb$l_bc_year = ucb->ucb$l_bc_year | (0x00FF & ucb->ucb$l_bc_register);
        };

        /* Place the data to be sent back in the QIO buffer in a local buffer */
        time_data.time            = time_value;
        time_data.planned         = ucb->ucb$l_bc_counter;
        time_data.unplanned       = ucb->ucb$l_bc_extra;
        time_data.mode            = ucb->ucb$l_bc_mode;
        time_data.format          = ucb->ucb$l_bc_format;
        time_data.year            = ucb->ucb$l_bc_year;
        time_data.sync            = ucb->ucb$l_bc_sync;
        time_data.n1              = ucb->ucb$l_bc_n1;
        time_data.n2              = ucb->ucb$l_bc_n2;
        time_data.code            = ucb->ucb$l_bc_code;
        time_data.modulation      = ucb->ucb$l_bc_modulation;
        time_data.offset          = ucb->ucb$l_bc_offset;
        time_data.utc             = ucb->ucb$l_bc_utc;
        time_data.leap_second     = ucb->ucb$l_bc_leap_second;
        time_data.leap_event      = ucb->ucb$l_bc_leap_event;
        time_data.leap_event_time = ucb->ucb$l_bc_leap_event_time;
        time_data.fw_major        = ucb->ucb$l_bc_fw_major;
        time_data.fw_minor        = ucb->ucb$l_bc_fw_minor;
        time_data.fw_month        = ucb->ucb$l_bc_fw_month;
        time_data.fw_day          = ucb->ucb$l_bc_fw_day;
        time_data.fw_year         = ucb->ucb$l_bc_fw_year;
        time_data.source          = ucb->ucb$l_bc_source;
        time_data.jam             = ucb->ucb$l_bc_jam;
        time_data.disciplining    = ucb->ucb$l_bc_disciplining;
        time_data.phase           = ucb->ucb$l_bc_phase;
        time_data.step            = ucb->ucb$l_bc_step;
        time_data.gain            = ucb->ucb$l_bc_gain;
        time_data.filter          = ucb->ucb$l_bc_filter;
        time_data.value           = ucb->ucb$l_bc_value;

        /* Move the contents of the local buffer to the user's QIO buffer */
        time_data_size = sizeof(time_data);
        ioc_std$movtouser(&time_data, time_data_size, (UCB *)ucb, &next_byte);

/*****************************************************************************/
/*****************************************************************************/

      }
      else
      {
        /* Set to return a normal status */
        status = SS$_NORMAL;

        /* Create the 64 bit time data information */
        time_value   = ucb->ucb$l_bc_time0;
        time_value   = time_value << 32;
        time_value_2 = ucb->ucb$l_bc_time1;
        time_value   = time_value | time_value_2;

        /* Place the data to be sent back in the QIO buffer in a local buffer */
        time_data.time            = time_value;
        time_data.planned         = ucb->ucb$l_bc_counter;
        time_data.unplanned       = ucb->ucb$l_bc_extra;
        time_data.mode            = ucb->ucb$l_bc_mode;
        time_data.format          = ucb->ucb$l_bc_format;
        time_data.year            = ucb->ucb$l_bc_year;
        time_data.sync            = ucb->ucb$l_bc_sync;
        time_data.n1              = ucb->ucb$l_bc_n1;
        time_data.n2              = ucb->ucb$l_bc_n2;
        time_data.code            = ucb->ucb$l_bc_code;
        time_data.modulation      = ucb->ucb$l_bc_modulation;
        time_data.offset          = ucb->ucb$l_bc_offset;
        time_data.utc             = ucb->ucb$l_bc_utc;
        time_data.leap_second     = ucb->ucb$l_bc_leap_second;
        time_data.leap_event      = ucb->ucb$l_bc_leap_event;
        time_data.leap_event_time = ucb->ucb$l_bc_leap_event_time;
        time_data.fw_major        = ucb->ucb$l_bc_fw_major;
        time_data.fw_minor        = ucb->ucb$l_bc_fw_minor;
        time_data.fw_month        = ucb->ucb$l_bc_fw_month;
        time_data.fw_day          = ucb->ucb$l_bc_fw_day;
        time_data.fw_year         = ucb->ucb$l_bc_fw_year;
        time_data.source          = ucb->ucb$l_bc_source;
        time_data.jam             = ucb->ucb$l_bc_jam;
        time_data.disciplining    = ucb->ucb$l_bc_disciplining;
        time_data.phase           = ucb->ucb$l_bc_phase;
        time_data.step            = ucb->ucb$l_bc_step;
        time_data.gain            = ucb->ucb$l_bc_gain;
        time_data.filter          = ucb->ucb$l_bc_filter;
        time_data.value           = ucb->ucb$l_bc_value;

        /* Move the contents of the local buffer to the user's QIO buffer */
        time_data_size = sizeof(time_data);
        ioc_std$movtouser(&time_data, time_data_size, (UCB *)ucb, &next_byte);

      };
    };

    /* Send this I/O request to I/O post processing */
    ioc_std$reqcom (status, 0, &(ucb->ucb$r_ucb));

    return;
}


/*
 * BC$WFI_TIMEOUT - Wait-for-interrupt timeout routine
 *
 * Functional description:
 *
 *   This routine is the wait-for-interrupt timeout routine.  It is called
 *   by exe$timeout when an operation set up by wfikpch takes more that the
 *   specified number of seconds.
 *
 *   Just cancels the IO
 *
 * Calling convention:
 *
 *   bc$wfi_timeout (irp, not_used, ucb)
 *
 * Input parameters:
 *
 *   irp        Pointer to I/O request packet
 *   not_used   Unused fork routine parameter fr4
 *   ucb        Pointer to unit control block
 *
 * Output parameters:
 *
 *   None.
 *
 * Return value:
 *
 *   None.
 *
 * Environment:
 * 
 *   Kernel mode, system context, device IPL, fork lock held, device lock held.
 *
 */

void bc$wfi_timeout (IRP *irp, void *not_used, BC_UCB *ucb) {

    /* Mark the request as timed out */
    ucb->ucb$r_ucb.ucb$v_timout = TRUE;

    /* Queue iodone processing */
    bc$iodone_fork (irp, 0, ucb);

    return;
}
